# -*- coding: utf-8 -*-

import redis


class CachedRedis(redis.Redis):
    def __init__(self, manager, opt_in, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self._manager = manager
        self._client_id = super().client_id()
        self.opt_in = opt_in
        try:
            if self.opt_in:
                self.execute_command(
                    "CLIENT",
                    "TRACKING",
                    "ON",
                    "redirect",
                    self._manager.client_id,
                    "OPTIN",
                )
            else:
                self.execute_command(
                    "CLIENT", "TRACKING", "ON", "redirect", self._manager.client_id
                )
        except redis.exceptions.ResponseError:
            raise NotImplementedError(
                "This redis version does not support this feature"
            )

    def close(self):
        self.execute_command("CLIENT", "TRACKING", "OFF")
        super().close()

    def get(self, name, opt_cache=False):
        try:
            value = self._manager.cache[name]
        except KeyError:
            if self.opt_in and opt_cache:
                self.__cache_next()
            value = super().get(name)
            if (self.opt_in and opt_cache) or not self.opt_in:
                self._manager.cache[name] = value
        except redis.exceptions.ConnectionError:
            print("reset called" * 20)
            self._manager.reset()
            raise
        return value

    def __cache_next(self):
        self.execute_command("CACHING", "NOREPLY")



# -*- coding: utf-8 -*-

from collections import OrderedDict


class LRUCache(OrderedDict):
    def __init__(self, manager, maxsize=256, *args, **kwargs):
        self.maxsize = maxsize
        self.manager = manager
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.manager.add(key)
        if len(self) > self.maxsize:
            oldest = next(iter(self))
            del self[oldest]
            self.manager.discard(oldest)



# -*- coding: utf-8 -*-

import redis
from collections import defaultdict
from .client import CachedRedis
from .pubssub_thread import LocalPubSubWorkerThread
from .cache import LRUCache
from .crc import crc64


class Manager(object):
    """ Redis Assisted Client Side Cache Manager """

    def __init__(
        self: object,
        pool: redis.ConnectionPool,
        capacity: int = 128,
        sleep_time: int = 0,
        opt_in=False,
    ):

        self.slots = None
        self.cache = None
        self.client_id = None
        self._thread = None
        self.pool = pool
        self.opt_in = opt_in
        self.capacity = capacity
        self.sleep_time = sleep_time
        self.client = redis.Redis(connection_pool=self.pool)
        self.reset()
        self.start()

    def __del__(self):
        if self.client_id is not None:
            self._thread.stop()
            self.client_id = None

    def reset(self):
        """ Resets the manager """
        self.slots = defaultdict(set)
        self.cache = LRUCache(self, maxsize=self.capacity)

    def start(self):
        """ Starts the manager """
        self.client_id = self.client.client_id()
        _pubsub = self.client.pubsub(ignore_subscribe_messages=True)
        _pubsub.subscribe(**{"__redis__:invalidate": self._handler})
        self._thread = LocalPubSubWorkerThread(
            self, _pubsub, sleep_time=self.sleep_time
        )
        self._thread.start()

    def _handler(self, message):
        """ Handles invalidation messages """
        slot = message["data"]
        self.invalidate(slot)

    def add(self, key):
        """ Adds a key to the internal tracking table """
        slot = self.slot(key)
        self.slots[slot].add(key)

    def discard(self, key):
        """ Removes a key from the internal tracking table """
        slot = self.slot(key)
        self.slots[slot].discard(key)

    def invalidate(self, slot):
        """ Invalidates a slot"s keys """
        slot = int(slot)
        while self.slots[slot]:
            key = self.slots[slot].pop()
            del self.cache[key]

    def get_connection(self, *args, **kwargs):
        """ Returns a cached Redis connection """
        conn = CachedRedis(
            self, connection_pool=self.pool, opt_in=self.opt_in, *args, **kwargs
        )
        return conn

    @staticmethod
    def slot(key):
        """ Returns the slot for a key """
        crc = crc64(key)
        crc &= 0xFFFFFF
        return crc



import threading
import redis


class LocalPubSubWorkerThread(threading.Thread):
    def __init__(self, manager, pubsub, sleep_time, daemon=False):
        super(LocalPubSubWorkerThread, self).__init__()
        self._manager = manager
        self.daemon = daemon
        self.pubsub = pubsub
        self.sleep_time = sleep_time
        self._running = threading.Event()

    def run(self):
        if self._running.is_set():
            return
        self._running.set()
        pubsub = self.pubsub
        sleep_time = self.sleep_time
        while self._running.is_set():
            try:
                pubsub.get_message(ignore_subscribe_messages=True, timeout=sleep_time)
            except redis.exceptions.ConnectionError:
                print("reset called" * 20)
                self._manager.reset()
                raise
        pubsub.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['redis_sacsc']

package_data = \
{'': ['*']}

install_requires = \
['redis>=3.4.1,<4.0.0']

setup_kwargs = {
    'name': 'redis-sacsc',
    'version': '0.0.1',
    'description': 'Redis Server-Assisted Client-Side Caching in Python',
    'long_description': '[![Build Status](https://travis-ci.com/iamajay/redis-sacsc.svg?branch=master)](https://travis-ci.com/iamajay/redis-sacsc)\n[![Coverage](https://coveralls.io/repos/github/iamajay/redis-sacsc/badge.svg?branch=master)](https://coveralls.io/github/iamajay/redis-sacsc?branch=master)\n\n# redis-sacsc\n\nRedis Server-Assisted Client-Side Caching in Python\n\n#### WARNING!\nThis feature is still in redis beta release, scheduled to be released in version 6\n\n\n## Features\n\n- Fully typed with annotations and checked with mypy, [PEP561 compatible](https://www.python.org/dev/peps/pep-0561/)\n- Redis Server-Assisted Client-Side Caching\n\n\n## Installation\n\n```bash\npip install redis-sacsc\n```\n\n\n## Example\n\nShowcase how your project can be used:\n\n```python\nimport redis\nfrom redis_sacsc import Manager\npool = redis.ConnectionPool.from_url("redis://127.0.0.1:6379")\nmanager = Manager(pool, capacity=512)\nredis_conn = manager.get_connection()\nredis_conn.set(\'foo\', \'bar\')\nredis_conn.get(\'foo\')\n# => \'bar\'\n\n```\n\n## License\n\n[MIT](https://github.com/iamajay/redis-sacsc/blob/master/LICENSE)\n',
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iamajay/redis-sacsc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
